@@ -1,18 +1,20 @@
 /**
  * @file activation.c
  *
+ * Copyright (c) 2016-2019 Nikias Bassen, All Rights Reserved.
+ * Copyright (c) 2014-2015 Martin Szulecki, All Rights Reserved.
  * Copyright (c) 2011-2014 Mirell Development, All Rights Reserved.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
- * 
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
@@ -40,11 +42,18 @@
 #endif
 #endif
 
+#ifdef WIN32
+#include <windows.h>
+#else
+#include <pthread.h>
+#endif
+
 #include <libideviceactivation.h>
 
 #define IDEVICE_ACTIVATION_USER_AGENT_IOS "iOS Device Activator (MobileActivation-20 built on Jan 15 2012 at 19:07:28)"
 #define IDEVICE_ACTIVATION_USER_AGENT_ITUNES "iTunes/11.1.4 (Macintosh; OS X 10.9.1) AppleWebKit/537.73.11"
 #define IDEVICE_ACTIVATION_DEFAULT_URL "https://albert.apple.com/deviceservices/deviceActivation"
+#define IDEVICE_ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL "https://albert.apple.com/deviceservices/drmHandshake"
 
 typedef enum {
 	IDEVICE_ACTIVATION_CONTENT_TYPE_URL_ENCODED,
@@ -69,62 +78,135 @@ struct idevice_activation_response_private {
 	char* title;
 	char* description;
 	plist_t activation_record;
+	plist_t headers;
 	plist_t fields;
 	plist_t fields_require_input;
+	plist_t fields_secure_input;
 	plist_t labels;
+	plist_t labels_placeholder;
 	int is_activation_ack;
 	int is_auth_required;
 	int has_errors;
 };
 
-int debug_level = 0;
+
+static void internal_libideviceactivation_init(void)
+{
+	curl_global_init(CURL_GLOBAL_ALL);
+}
+
+static void internal_libideviceactivation_deinit(void)
+{
+	curl_global_cleanup();
+}
+
+#ifdef WIN32
+
+typedef volatile struct {
+    LONG lock;
+    int state;
+} thread_once_t;
+
+static thread_once_t init_once = {0, 0};
+static thread_once_t deinit_once = {0, 0};
+
+static void thread_once(thread_once_t *once_control, void (*init_routine)(void))
+{
+    while (InterlockedExchange(&(once_control->lock), 1) != 0) {
+        Sleep(1);
+    }
+    if (!once_control->state) {
+        once_control->state = 1;
+        init_routine();
+    }
+    InterlockedExchange(&(once_control->lock), 0);
+}
+
+BOOL WINAPI DllMain(HINSTANCE hModule, DWORD dwReason, LPVOID lpReserved)
+{
+    switch (dwReason) {
+    case DLL_PROCESS_ATTACH:
+        thread_once(&init_once, internal_libideviceactivation_init);
+        break;
+    case DLL_PROCESS_DETACH:
+        thread_once(&deinit_once, internal_libideviceactivation_deinit);
+        break;
+    default:
+        break;
+    }
+    return 1;
+}
+
+#else
+
+static pthread_once_t init_once = PTHREAD_ONCE_INIT;
+static pthread_once_t deinit_once = PTHREAD_ONCE_INIT;
+
+static void __attribute__((constructor)) libideviceactivation_initialize(void)
+{
+    pthread_once(&init_once, internal_libideviceactivation_init);
+}
+
+static void __attribute__((destructor)) libideviceactivation_deinitialize(void)
+{
+    pthread_once(&deinit_once, internal_libideviceactivation_deinit);
+}
+
+#endif
+
+
+static int debug_level = 0;
 
 IDEVICE_ACTIVATION_API void idevice_activation_set_debug_level(int level) {
 	debug_level = level;
 }
 
-static idevice_activation_error_t idevice_activation_activation_node_from_plist_xml(const char* plist_xml, size_t size, plist_t* activation_node)
+static idevice_activation_error_t idevice_activation_activation_record_from_plist(idevice_activation_response_t response, plist_t plist)
 {
-	plist_t activation_ticket = NULL;
-
-	plist_from_xml(plist_xml, size, &activation_ticket);
-	if (activation_ticket == NULL) {
-		return IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
-	}
-
-	plist_t activation_node_tmp = plist_dict_get_item(activation_ticket, "iphone-activation");
-	if (!activation_node_tmp) {
-		activation_node_tmp = plist_dict_get_item(activation_ticket, "device-activation");
-		if (!activation_node_tmp) {
-			plist_free(activation_ticket);
+	plist_t record = plist_dict_get_item(plist, "ActivationRecord");
+	if (record != NULL) {
+		plist_t ack_received = plist_dict_get_item(record, "ack-received");
+		if (ack_received) {
+			uint8_t val = 0;
+			plist_get_bool_val(ack_received, &val);
+			if (val) {
+				response->is_activation_ack = 1;
+			}
+		}
+		response->activation_record = plist_new_data(response->raw_content, response->raw_content_size);
+	} else {
+		plist_t activation_node = plist_dict_get_item(plist, "iphone-activation");
+		if (!activation_node) {
+			activation_node = plist_dict_get_item(plist, "device-activation");
+		}
+		if (!activation_node) {
 			return IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
 		}
+		plist_t ack_received = plist_dict_get_item(activation_node, "ack-received");
+		if (ack_received) {
+			uint8_t val = 0;
+			plist_get_bool_val(ack_received, &val);
+			if (val) {
+				response->is_activation_ack = 1;
+			}
+		}
+		record = plist_dict_get_item(activation_node, "activation-record");
+		if (record) {
+			response->activation_record = plist_copy(record);
+		}
 	}
-
-	*activation_node = plist_copy(activation_node_tmp);
-	plist_free(activation_ticket);
-
 	return IDEVICE_ACTIVATION_E_SUCCESS;
 }
 
-static idevice_activation_error_t idevice_activation_activation_record_from_activation_node(plist_t activation_node, plist_t* activation_record)
-{
-	plist_t record = plist_dict_get_item(activation_node, "activation-record");
-	if (!record) {
-		return IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
-	}
-
-	*activation_record = plist_copy(record);
-
-	return IDEVICE_ACTIVATION_E_SUCCESS;
-}
-
-static void idevice_activation_response_add_field(idevice_activation_response_t response, const char* key, const char* value, int required_input)
+static void idevice_activation_response_add_field(idevice_activation_response_t response, const char* key, const char* value, int required_input, int secure_input)
 {
 	plist_dict_set_item(response->fields, key, plist_new_string(value));
 	if (required_input) {
 		plist_dict_set_item(response->fields_require_input, key, plist_new_bool(1));
 	}
+	if (secure_input) {
+		plist_dict_set_item(response->fields_secure_input, key, plist_new_bool(1));
+	}
 }
 
 static idevice_activation_error_t idevice_activation_parse_buddyml_response(idevice_activation_response_t response)
@@ -135,7 +217,7 @@ static idevice_activation_error_t idevice_activation_parse_buddyml_response(idev
 	xmlXPathObjectPtr xpath_result = NULL;
 	int i = 0;
 
-	if (!response->content_type == IDEVICE_ACTIVATION_CONTENT_TYPE_BUDDYML)
+	if (response->content_type != IDEVICE_ACTIVATION_CONTENT_TYPE_BUDDYML)
 		return IDEVICE_ACTIVATION_E_UNKNOWN_CONTENT_TYPE;
 
 	doc = xmlReadMemory(response->raw_content, response->raw_content_size, "ideviceactivation.xml", NULL, XML_PARSE_NOERROR);
@@ -232,11 +314,17 @@ static idevice_activation_error_t idevice_activation_parse_buddyml_response(idev
 		xmlXPathFreeObject(xpath_result);
 		xpath_result = NULL;
 	}
-	xpath_result = xmlXPathEvalExpression((const xmlChar*) "/xmlui/page/tableView/section[@footer and not(@footerLinkURL)]/@footer", context);
+	xpath_result = xmlXPathEvalExpression((const xmlChar*) "/xmlui/page/tableView/section/footer[not (@url)]", context);
 	if (!xpath_result) {
 		result = IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
 		goto cleanup;
 	}
+	if (!xpath_result->nodesetval) {
+		xmlXPathFreeObject(xpath_result);
+		xpath_result = xmlXPathEvalExpression((const xmlChar*) "/xmlui/page/tableView/section[@footer and not(@footerLinkURL)]/@footer", context);
+		result = IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+		goto cleanup;
+	}
 
 	if (xpath_result->nodesetval) {
 		char* response_description = (char*) malloc(sizeof(char));
@@ -266,6 +354,7 @@ static idevice_activation_error_t idevice_activation_parse_buddyml_response(idev
 		xmlXPathFreeObject(xpath_result);
 		xpath_result = NULL;
 	}
+
 	xpath_result = xmlXPathEvalExpression((const xmlChar*) "/xmlui/page//editableTextRow", context);
 	if (!xpath_result) {
 		result = IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
@@ -279,14 +368,27 @@ static idevice_activation_error_t idevice_activation_parse_buddyml_response(idev
 				result = IDEVICE_ACTIVATION_E_BUDDYML_PARSING_ERROR;
 				goto cleanup;
 			}
+			int secure_input = 0;
+			xmlChar* secure = xmlGetProp(xpath_result->nodesetval->nodeTab[i], (const xmlChar*) "secure");
+			if (secure) {
+				if (!strcmp((const char*)secure, "true")) {
+					secure_input = 1;
+				}
+				xmlFree(secure);
+			}
 
-			idevice_activation_response_add_field(response, (const char*) id, "", 1);
+			idevice_activation_response_add_field(response, (const char*) id, "", 1, secure_input);
 
 			xmlChar* label = xmlGetProp(xpath_result->nodesetval->nodeTab[i], (const xmlChar*) "label");
 			if (label) {
 				plist_dict_set_item(response->labels, (const char*)id, plist_new_string((const char*) label));
 				xmlFree(label);
 			}
+			xmlChar* placeholder = xmlGetProp(xpath_result->nodesetval->nodeTab[i], (const xmlChar*) "placeholder");
+			if (placeholder) {
+				plist_dict_set_item(response->labels_placeholder, (const char*)id, plist_new_string((const char*) placeholder));
+				xmlFree(placeholder);
+			}
 
 			xmlFree(id);
 		}
@@ -313,7 +415,7 @@ static idevice_activation_error_t idevice_activation_parse_buddyml_response(idev
 				}
 
 				idevice_activation_response_add_field(response,
-					(const char*) xpath_result->nodesetval->nodeTab[i]->name, (const char*) content, 0);
+					(const char*) xpath_result->nodesetval->nodeTab[i]->name, (const char*) content, 0, 0);
 				xmlFree(content);
 			}
 		}
@@ -341,7 +443,7 @@ static idevice_activation_error_t idevice_activation_parse_html_response(idevice
 	xmlXPathContextPtr context = NULL;
 	xmlXPathObjectPtr xpath_result = NULL;
 
-	if (!response->content_type == IDEVICE_ACTIVATION_CONTENT_TYPE_HTML)
+	if (response->content_type != IDEVICE_ACTIVATION_CONTENT_TYPE_HTML)
 		return IDEVICE_ACTIVATION_E_UNKNOWN_CONTENT_TYPE;
 
 	doc = xmlReadMemory(response->raw_content, response->raw_content_size, "ideviceactivation.xml", NULL, XML_PARSE_RECOVER | XML_PARSE_NOERROR);
@@ -380,46 +482,24 @@ static idevice_activation_error_t idevice_activation_parse_html_response(idevice
 	}
 
 	if (xpath_result->nodesetval && xpath_result->nodesetval->nodeNr) {
-		plist_t activation_node = NULL;
+		plist_t plist = NULL;
 		xmlBufferPtr plistNodeBuffer = xmlBufferCreate();
 		if (htmlNodeDump(plistNodeBuffer, doc, xpath_result->nodesetval->nodeTab[0]) == -1) {
 			result = IDEVICE_ACTIVATION_E_HTML_PARSING_ERROR;
 			goto local_cleanup;
 		}
 
-		result = idevice_activation_activation_node_from_plist_xml(
-			(const char*) plistNodeBuffer->content, plistNodeBuffer->use, &activation_node);
-		if (result != IDEVICE_ACTIVATION_E_SUCCESS) {
-			response->has_errors = 1;
-			result = IDEVICE_ACTIVATION_E_SUCCESS;
-			goto local_cleanup;
-		}
-
-		// check for ack-received
-		plist_t ack_received = plist_dict_get_item(activation_node, "ack-received");
-		if (ack_received) {
-			uint8_t val = 0;
-			plist_get_bool_val(ack_received, &val);
-			if (val) {
-				response->is_activation_ack = 1;
-			} else {
-				result = IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
-			}
-
-			goto local_cleanup;
-		}
-
-		// try to retrieve the activation record
-		result = idevice_activation_activation_record_from_activation_node(activation_node, &response->activation_record);
-		if (result != IDEVICE_ACTIVATION_E_SUCCESS) {
+		plist_from_xml((const char*) plistNodeBuffer->content, plistNodeBuffer->use, &plist);
+		if (!plist) {
+			result = IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
 			goto local_cleanup;
 		}
+		result = idevice_activation_activation_record_from_plist(response, plist);
+		plist_free(plist);
 
 	local_cleanup:
 		if (plistNodeBuffer)
 			xmlBufferFree(plistNodeBuffer);
-		if (activation_node)
-			plist_free(activation_node);
 		goto cleanup;
 	}
 
@@ -442,16 +522,25 @@ static idevice_activation_error_t idevice_activation_parse_raw_response(idevice_
 	{
 		case IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST:
 		{
-			plist_t activation_node = NULL;
 			idevice_activation_error_t result = IDEVICE_ACTIVATION_E_SUCCESS;
 
-			result = idevice_activation_activation_node_from_plist_xml(response->raw_content, response->raw_content_size, &activation_node);
-			if (result != IDEVICE_ACTIVATION_E_SUCCESS) {
-				return result;
+			plist_t plist = NULL;
+			plist_from_xml(response->raw_content, response->raw_content_size, &plist);
+
+			if (plist == NULL) {
+				return IDEVICE_ACTIVATION_E_PLIST_PARSING_ERROR;
 			}
 
-			result = idevice_activation_activation_record_from_activation_node(activation_node, &response->activation_record);
-			plist_free(activation_node);
+			/* check if this is a reply to drmHandshake request */
+			if (plist_dict_get_item(plist, "HandshakeResponseMessage") != NULL) {
+				result = IDEVICE_ACTIVATION_E_SUCCESS;
+			} else {
+				result = idevice_activation_activation_record_from_plist(response, plist);
+			}
+
+			plist_free(response->fields);
+			response->fields = plist;
+
 			return result;
 		}
 		case IDEVICE_ACTIVATION_CONTENT_TYPE_BUDDYML:
@@ -471,8 +560,9 @@ static size_t idevice_activation_write_callback(char* data, size_t size, size_t
 	const size_t total = size * nmemb;
 
 	if (total != 0) {
-		response->raw_content = realloc(response->raw_content, response->raw_content_size + total);
+		response->raw_content = realloc(response->raw_content, response->raw_content_size + total + 1);
 		memcpy(response->raw_content + response->raw_content_size, data, total);
+		response->raw_content[response->raw_content_size + total] = '\0';
 		response->raw_content_size += total;
 	}
 
@@ -484,13 +574,42 @@ static size_t idevice_activation_header_callback(void *data, size_t size, size_t
 	idevice_activation_response_t response = (idevice_activation_response_t)userdata;
 	const size_t total = size * nmemb;
 	if (total != 0) {
-		if (strstr(data, "Content-Type: text/xml") != NULL) {
-			response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST;
-		} else if (strstr(data, "Content-Type: application/x-buddyml") != NULL) {
-			response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_BUDDYML;
-		} else if (strstr(data, "Content-Type: text/html") != NULL) {
-			response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_HTML;
+		char *header = malloc(total + 1);
+		char *value = NULL;
+		char *p = NULL;
+		memcpy(header, data, total);
+		header[total] = '\0';
+
+		p = strchr(header, ':');
+		if (p) {
+			*p = '\0';
+			p++;
+			while (*p == ' ') {
+				p++;
+			}
+			if (*p != '\0') {
+				value = p;
+				while (*p != '\0' && *p != '\n' && *p != '\r') {
+					p++;
+				}
+				*p = '\0';
+			}
 		}
+		if (value) {
+			if (strcmp(header, "Content-Type") == 0) {
+				if (strcmp(value, "text/xml") == 0) {
+					response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST;
+				} else if (strcmp(value, "application/xml") == 0) {
+					response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST;
+				} else if (strcmp(value, "application/x-buddyml") == 0) {
+					response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_BUDDYML;
+				} else if (strcmp(value, "text/html") == 0) {
+					response->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_HTML;
+				}
+			}
+			plist_dict_set_item(response->headers, header, plist_new_string(value));
+		}
+		free(header);
 	}
 	return total;
 }
@@ -542,35 +661,64 @@ static char* urlencode(const char* buf)
 
 static int plist_strip_xml(char** xmlplist)
 {
-	uint32_t size = 0;
-
-	if (!xmlplist && !*xmlplist)
+	if (!xmlplist || !*xmlplist)
 		return -1;
 
 	char* start = strstr(*xmlplist, "<plist version=\"1.0\">\n");
-	if (start == NULL) {
+	if (!start)
 		return -1;
-	}
 
 	char* stop = strstr(*xmlplist, "\n</plist>");
-	if (stop == NULL) {
+	if (!stop)
 		return -1;
-	}
 
 	start += strlen("<plist version=\"1.0\">\n");
-	size = stop - start;
+	uint32_t size = stop - start;
 	char* stripped = malloc(size + 1);
-	memset(stripped, '\0', size + 1);
+	if (!stripped)
+		return -1;
+
 	memcpy(stripped, start, size);
+	stripped[size] = '\0';
 	free(*xmlplist);
 	*xmlplist = stripped;
-	stripped = NULL;
 
 	return 0;
 }
 
+static int idevice_activation_curl_debug_callback(CURL *handle, curl_infotype type, char *data, size_t size, void *userptr)
+{
+	switch (type) {
+	case CURLINFO_TEXT:
+		fprintf(stderr, "* ");
+		break;
+	case CURLINFO_DATA_IN:
+	case CURLINFO_HEADER_IN:
+		fprintf(stderr, "< ");
+		break;
+	case CURLINFO_DATA_OUT:
+	case CURLINFO_HEADER_OUT:
+		fprintf(stderr, "> ");
+		break;
+	case CURLINFO_SSL_DATA_IN:
+	case CURLINFO_SSL_DATA_OUT:
+		return 0;
+	default:
+		return 0;
+	}
+
+	fwrite(data, 1, size, stderr);
+	if (size > 0 && data[size-1] != '\n') {
+		fprintf(stderr, "\n");
+	}
+	return 0;
+}
+
 IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_request_new(idevice_activation_client_type_t client_type, idevice_activation_request_t* request)
 {
+	if (!request)
+		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+
 	idevice_activation_request_t tmp_request = (idevice_activation_request_t) malloc(sizeof(idevice_activation_request));
 
 	if (!tmp_request) {
@@ -588,76 +736,103 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_request_new
 
 IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_request_new_from_lockdownd(idevice_activation_client_type_t client_type, lockdownd_client_t lockdown, idevice_activation_request** request)
 {
-	// check arguments
-	if (!lockdown) {
+	if (!lockdown || !request) {
 		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
 	}
 
+	uint8_t has_telephony_capability = 0;
+	uint8_t has_mobile_equipment_id = 0;
+	lockdownd_error_t err;
+	plist_t info = NULL;
 	plist_t node = NULL;
 	plist_t fields = plist_new_dict();
 
 	// add InStoreActivation
 	plist_dict_set_item(fields, "InStoreActivation", plist_new_string("false"));
 
-	// add AppleSerialNumber
-	if ((lockdownd_get_value(lockdown, NULL, "SerialNumber", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_STRING)) {
-		fprintf(stderr, "%s: Unable to get SerialNumber from lockdownd\n", __func__);
+	// get a bunch of information at once
+	err = lockdownd_get_value(lockdown, NULL, NULL, &info);
+	if (err != LOCKDOWN_E_SUCCESS) {
+		if (debug_level > 0)
+			fprintf(stderr, "%s: Unable to get basic information from lockdownd\n", __func__);
 		plist_free(fields);
 		return IDEVICE_ACTIVATION_E_INCOMPLETE_INFO;
-	} else {
-		plist_dict_set_item(fields, "AppleSerialNumber", plist_copy(node));
-	}
-	if (node) {
-		plist_free(node);
-		node = NULL;
 	}
 
-	// add IMEI
-	if ((lockdownd_get_value(lockdown, NULL, "InternationalMobileEquipmentIdentity", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_STRING)) {
-		fprintf(stderr, "%s: Unable to get IMEI from lockdownd\n", __func__);
-	} else {
-		plist_dict_set_item(fields, "IMEI", plist_copy(node));
-	}
-	if (node) {
-		plist_free(node);
-		node = NULL;
+	// add AppleSerialNumber
+	node = plist_dict_get_item(info, "SerialNumber");
+	if (!node || plist_get_node_type(node) != PLIST_STRING) {
+		if (debug_level > 0)
+			fprintf(stderr, "%s: Unable to get SerialNumber from lockdownd\n", __func__);
+		plist_free(fields);
+		plist_free(info);
+		return IDEVICE_ACTIVATION_E_INCOMPLETE_INFO;
 	}
+	plist_dict_set_item(fields, "AppleSerialNumber", plist_copy(node));
+	node = NULL;
 
-	// add MEID
-	if ((lockdownd_get_value(lockdown, NULL, "MobileEquipmentIdentifier", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_STRING)) {
-		fprintf(stderr, "%s: Unable to get MEID from lockdownd\n", __func__);
-	} else {
-		plist_dict_set_item(fields, "MEID", plist_copy(node));
-	}
-	if (node) {
-		plist_free(node);
-		node = NULL;
-	}
 
-	// add IMSI
-	if ((lockdownd_get_value(lockdown, NULL, "InternationalMobileSubscriberIdentity", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_STRING)) {
-		fprintf(stderr, "%s: Unable to get IMSI from lockdownd\n", __func__);
+	// check if device has telephone capability
+	node = plist_dict_get_item(info, "TelephonyCapability");
+	if (!node || plist_get_node_type(node) != PLIST_BOOLEAN) {
+		has_telephony_capability = 0;
 	} else {
-		plist_dict_set_item(fields, "IMSI", plist_copy(node));
-	}
-	if (node) {
-		plist_free(node);
-		node = NULL;
+		plist_get_bool_val(node, &has_telephony_capability);
 	}
+	node = NULL;
 
-	// add ICCID
-	if ((lockdownd_get_value(lockdown, NULL, "IntegratedCircuitCardIdentity", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_STRING)) {
-		fprintf(stderr, "%s: Unable to get ICCID from lockdownd\n", __func__);
-	} else {
-		plist_dict_set_item(fields, "ICCID", plist_copy(node));
-	}
-	if (node) {
-		plist_free(node);
+	if (has_telephony_capability) {
+		// add IMEI
+		node = plist_dict_get_item(info, "InternationalMobileEquipmentIdentity");
+		if (!node || plist_get_node_type(node) != PLIST_STRING) {
+			has_mobile_equipment_id = 0;
+		} else {
+			plist_dict_set_item(fields, "IMEI", plist_copy(node));
+			has_mobile_equipment_id = 1;
+		}
+		node = NULL;
+
+		// add MEID
+		node = plist_dict_get_item(info, "MobileEquipmentIdentifier");
+		if (!node || plist_get_node_type(node) != PLIST_STRING) {
+			if (debug_level > 0)
+				fprintf(stderr, "%s: Unable to get MEID from lockdownd\n", __func__);
+			if (!has_mobile_equipment_id) {
+				plist_free(fields);
+				plist_free(info);
+				return IDEVICE_ACTIVATION_E_INCOMPLETE_INFO;
+			}
+		} else {
+			plist_dict_set_item(fields, "MEID", plist_copy(node));
+		}
+		node = NULL;
+
+		// add IMSI
+		node = plist_dict_get_item(info, "InternationalMobileSubscriberIdentity");
+		if (!node || plist_get_node_type(node) != PLIST_STRING) {
+			if (debug_level > 0)
+				fprintf(stderr, "%s: Unable to get IMSI from lockdownd\n", __func__);
+		} else {
+			plist_dict_set_item(fields, "IMSI", plist_copy(node));
+		}
+		node = NULL;
+
+		// add ICCID
+		node = plist_dict_get_item(info, "IntegratedCircuitCardIdentity");
+		if (!node || plist_get_node_type(node) != PLIST_STRING) {
+			if (debug_level > 0)
+				fprintf(stderr, "%s: Unable to get ICCID from lockdownd\n", __func__);
+		} else {
+			plist_dict_set_item(fields, "ICCID", plist_copy(node));
+		}
 		node = NULL;
 	}
+	plist_free(info);
+	info = NULL;
 
 	// add activation-info
-	if ((lockdownd_get_value(lockdown, NULL, "ActivationInfo", &node) != LOCKDOWN_E_SUCCESS) || !node || (plist_get_node_type(node) != PLIST_DICT)) {
+	err = lockdownd_get_value(lockdown, NULL, "ActivationInfo", &node);
+	if (err != LOCKDOWN_E_SUCCESS || !node || plist_get_node_type(node) != PLIST_DICT) {
 		fprintf(stderr, "%s: Unable to get ActivationInfo from lockdownd\n", __func__);
 		plist_free(fields);
 		return IDEVICE_ACTIVATION_E_INCOMPLETE_INFO;
@@ -682,25 +857,46 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_request_new
 	return IDEVICE_ACTIVATION_E_SUCCESS;
 }
 
+IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_drm_handshake_request_new(idevice_activation_client_type_t client_type, idevice_activation_request_t* request)
+{
+	if (!request)
+		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+
+	idevice_activation_request_t tmp_request = (idevice_activation_request_t) malloc(sizeof(idevice_activation_request));
+
+	if (!tmp_request) {
+		return IDEVICE_ACTIVATION_E_OUT_OF_MEMORY;
+	}
+
+	tmp_request->client_type = client_type;
+	tmp_request->content_type = IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST;
+	tmp_request->url = strdup(IDEVICE_ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL);
+	tmp_request->fields = plist_new_dict();
+	*request = tmp_request;
+
+	return IDEVICE_ACTIVATION_E_SUCCESS;
+}
+
 IDEVICE_ACTIVATION_API void idevice_activation_request_free(idevice_activation_request_t request)
 {
 	if (!request)
 		return;
 
-	if (request->fields)
-		plist_free(request->fields);
-
+	plist_free(request->fields);
 	free(request);
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_get_fields(idevice_activation_request_t request, plist_t* fields)
 {
+	if (!request || !fields)
+		return;
+
 	*fields = plist_copy(request->fields);
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_set_fields(idevice_activation_request_t request, plist_t fields)
 {
-	if (!fields)
+	if (!request || !fields)
 		return;
 
 	if (request->content_type == IDEVICE_ACTIVATION_CONTENT_TYPE_URL_ENCODED) {
@@ -722,6 +918,9 @@ IDEVICE_ACTIVATION_API void idevice_activation_request_set_fields(idevice_activa
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_set_fields_from_response(idevice_activation_request_t request, const idevice_activation_response_t response)
 {
+	if (!request || !response)
+		return;
+
 	plist_t response_fields = NULL;
 	idevice_activation_response_get_fields(response, &response_fields);
 	if (response_fields) {
@@ -732,11 +931,17 @@ IDEVICE_ACTIVATION_API void idevice_activation_request_set_fields_from_response(
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_set_field(idevice_activation_request_t request, const char* key, const char* value)
 {
+	if (!request || !key || !value)
+		return;
+
 	plist_dict_set_item(request->fields, key, plist_new_string(value));
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_get_field(idevice_activation_request_t request, const char* key, char** value)
 {
+	if (!request || !key || !value)
+		return;
+
 	char* tmp_value = NULL;
 
 	plist_t item = plist_dict_get_item(request->fields, key);
@@ -754,20 +959,26 @@ IDEVICE_ACTIVATION_API void idevice_activation_request_get_field(idevice_activat
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_get_url(idevice_activation_request_t request, const char** url)
 {
+	if (!request || !url)
+		return;
+
 	*url = request->url;
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_request_set_url(idevice_activation_request_t request, const char* url)
 {
-	if (request->url) {
-		free(request->url);
-	}
+	if (!request || !url)
+		return;
 
+	free(request->url);
 	request->url = strdup(url);
 }
 
 IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_new(idevice_activation_response_t* response)
 {
+	if (!response)
+		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+
 	idevice_activation_response_t tmp_response = (idevice_activation_response_t) malloc(sizeof(idevice_activation_response));
 
 	if (!tmp_response) {
@@ -780,9 +991,12 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_ne
 	tmp_response->title = NULL;
 	tmp_response->description = NULL;
 	tmp_response->activation_record = NULL;
+	tmp_response->headers = plist_new_dict();
 	tmp_response->fields = plist_new_dict();
 	tmp_response->fields_require_input = plist_new_dict();
+	tmp_response->fields_secure_input = plist_new_dict();
 	tmp_response->labels = plist_new_dict();
+	tmp_response->labels_placeholder = plist_new_dict();
 	tmp_response->is_activation_ack = 0;
 	tmp_response->is_auth_required = 0;
 	tmp_response->has_errors = 0;
@@ -793,6 +1007,9 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_ne
 
 IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_new_from_html(const char* content, idevice_activation_response_t* response)
 {
+	if (!content || !response)
+		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+
 	idevice_activation_response_t tmp_response = NULL;
 	idevice_activation_error_t result = IDEVICE_ACTIVATION_E_SUCCESS;
 
@@ -828,6 +1045,9 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_ne
 
 IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_to_buffer(idevice_activation_response_t response, char** buffer, size_t* size)
 {
+	if (!response || !buffer || !size)
+		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
+
 	char* tmp_buffer = (char*) malloc(sizeof(char) * response->raw_content_size);
 	if (!tmp_buffer) {
 		return IDEVICE_ACTIVATION_E_OUT_OF_MEMORY;
@@ -843,25 +1063,27 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_response_to
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_free(idevice_activation_response_t response)
 {
-	if (response->raw_content)
-		free(response->raw_content);
-	if (response->title)
-		free(response->title);
-	if (response->description)
-		free(response->description);
-	if (response->activation_record)
-		plist_free(response->activation_record);
-	if (response->fields)
-		plist_free(response->fields);
-	if (response->fields_require_input)
-		plist_free(response->fields_require_input);
-	if (response->labels)
-		plist_free(response->labels);
+	if (!response)
+		return;
+
+	free(response->raw_content);
+	free(response->title);
+	free(response->description);
+	plist_free(response->activation_record);
+	plist_free(response->headers);
+	plist_free(response->fields);
+	plist_free(response->fields_require_input);
+	plist_free(response->fields_secure_input);
+	plist_free(response->labels);
+	plist_free(response->labels_placeholder);
 	free(response);
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_field(idevice_activation_response_t response, const char* key, char** value)
 {
+	if (!response || !key || !value)
+		return;
+
 	*value = NULL;
 	plist_t item = plist_dict_get_item(response->fields, key);
 
@@ -872,11 +1094,16 @@ IDEVICE_ACTIVATION_API void idevice_activation_response_get_field(idevice_activa
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_fields(idevice_activation_response_t response, plist_t* fields)
 {
-	*fields = plist_copy(response->fields);
+	if (response && response->fields && fields) {
+		*fields = plist_copy(response->fields);
+	}
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_label(idevice_activation_response_t response, const char* key, char** value)
 {
+	if (!response || !key || !value)
+		return;
+
 	*value = NULL;
 	plist_t item = plist_dict_get_item(response->labels, key);
 	if (item) {
@@ -884,18 +1111,39 @@ IDEVICE_ACTIVATION_API void idevice_activation_response_get_label(idevice_activa
 	}
 }
 
+IDEVICE_ACTIVATION_API void idevice_activation_response_get_placeholder(idevice_activation_response_t response, const char* key, char** value)
+{
+	if (!response || !key || !value)
+		return;
+
+	*value = NULL;
+	plist_t item = plist_dict_get_item(response->labels_placeholder, key);
+	if (item) {
+		plist_get_string_val(item, value);
+	}
+}
+
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_title(idevice_activation_response_t response, const char** title)
 {
+	if (!response || !title)
+		return;
+
 	*title = response->title;
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_description(idevice_activation_response_t response, const char** description)
 {
+	if (!response || !description)
+		return;
+
 	*description = response->description;
 }
 
 IDEVICE_ACTIVATION_API void idevice_activation_response_get_activation_record(idevice_activation_response_t response, plist_t* activation_record)
 {
+	if (!response || !activation_record)
+		return;
+
 	if (response->activation_record) {
 		*activation_record = plist_copy(response->activation_record);
 	} else {
@@ -903,26 +1151,51 @@ IDEVICE_ACTIVATION_API void idevice_activation_response_get_activation_record(id
 	}
 }
 
+IDEVICE_ACTIVATION_API void idevice_activation_response_get_headers(idevice_activation_response_t response, plist_t* headers)
+{
+	if (!response || !headers)
+		return;
+
+	*headers = plist_copy(response->headers);
+}
+
 IDEVICE_ACTIVATION_API int idevice_activation_response_is_activation_acknowledged(idevice_activation_response_t response)
 {
+	if (!response)
+		return 0;
+
 	return response->is_activation_ack;
 }
 
 IDEVICE_ACTIVATION_API int idevice_activation_response_is_authentication_required(idevice_activation_response_t response)
 {
+	if (!response)
+		return 0;
+
 	return response->is_auth_required;
 }
 
 IDEVICE_ACTIVATION_API int idevice_activation_response_field_requires_input(idevice_activation_response_t response, const char* key)
 {
-	if (plist_dict_get_item(response->fields_require_input, key)) {
-		return 1;
-	}
-	return 0;
+	if (!response || !key)
+		return 0;
+
+	return (plist_dict_get_item(response->fields_require_input, key) ? 1 : 0);
+}
+
+IDEVICE_ACTIVATION_API int idevice_activation_response_field_secure_input(idevice_activation_response_t response, const char* key)
+{
+	if (!response || !key)
+		return 0;
+
+	return (plist_dict_get_item(response->fields_secure_input, key) ? 1 : 0);
 }
 
 IDEVICE_ACTIVATION_API int idevice_activation_response_has_errors(idevice_activation_response_t response)
 {
+	if (!response)
+		return 0;
+
 	return response->has_errors;
 }
 
@@ -931,7 +1204,7 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 	idevice_activation_error_t result = IDEVICE_ACTIVATION_E_SUCCESS;
 
 	// check arguments
-	if (!request) {
+	if (!request || !response) {
 		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
 	}
 
@@ -939,12 +1212,11 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 	plist_dict_new_iter(request->fields, &iter);
 	if (!iter) {
 		return IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
-		goto cleanup;
 	}
 
-	curl_global_init(CURL_GLOBAL_ALL);
 	CURL* handle = curl_easy_init();
 	struct curl_httppost* form = NULL;
+	struct curl_slist* slist = NULL;
 
 	if (!handle) {
 		result = IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
@@ -984,8 +1256,7 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 
 					curl_formadd(&form, &last, CURLFORM_COPYNAME, key, CURLFORM_COPYCONTENTS, svalue, CURLFORM_END);
 
-					if (svalue)
-						free(svalue);
+					free(svalue);
 					svalue = NULL;
 				}
 			}
@@ -1004,8 +1275,7 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 						plist_get_string_val(value_node, &svalue);
 					} else {
 						// only strings supported
-						if (postdata)
-							free(postdata);
+						free(postdata);
 						result = IDEVICE_ACTIVATION_E_UNSUPPORTED_FIELD_TYPE;
 						goto cleanup;
 					}
@@ -1018,8 +1288,7 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 						free(value_encoded);
 					}
 
-					if (svalue)
-						free(svalue);
+					free(svalue);
 					svalue = NULL;
 				}
 			}
@@ -1033,6 +1302,16 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 		curl_easy_setopt(handle, CURLOPT_POST, 1);
 		curl_easy_setopt(handle, CURLOPT_POSTFIELDS, postdata);
 		curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, strlen(postdata));
+	} else if (request->content_type == IDEVICE_ACTIVATION_CONTENT_TYPE_PLIST) {
+		char *postdata = NULL;
+		uint32_t postdata_len = 0;
+		plist_to_xml(request->fields, &postdata, &postdata_len);
+		curl_easy_setopt(handle, CURLOPT_POST, 1);
+		curl_easy_setopt(handle, CURLOPT_POSTFIELDS, postdata);
+		curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, postdata_len);
+		slist = curl_slist_append(NULL, "Content-Type: application/x-apple-plist");
+		slist = curl_slist_append(slist, "Accept: application/xml");
+		curl_easy_setopt(handle, CURLOPT_HTTPHEADER, slist);
 	}
 	else {
 		result = IDEVICE_ACTIVATION_E_INTERNAL_ERROR;
@@ -1051,18 +1330,17 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 	curl_easy_setopt(handle, CURLOPT_HEADERDATA, tmp_response);
 	curl_easy_setopt(handle, CURLOPT_HEADERFUNCTION, &idevice_activation_header_callback);
 	curl_easy_setopt(handle, CURLOPT_URL, request->url);
+	curl_easy_setopt(handle, CURLOPT_FORBID_REUSE, 1);
+	curl_easy_setopt(handle, CURLOPT_FOLLOWLOCATION, 1L);
 
 	// enable communication debugging
 	if (debug_level > 0) {
 		curl_easy_setopt(handle, CURLOPT_VERBOSE, 1);
+		curl_easy_setopt(handle, CURLOPT_DEBUGFUNCTION, idevice_activation_curl_debug_callback);
 	}
 
 	curl_easy_perform(handle);
 
-	if (debug_level > 0) {
-		fprintf(stderr, "%*s\n", (int)tmp_response->raw_content_size, tmp_response->raw_content);
-	}
-
 	result = idevice_activation_parse_raw_response(tmp_response);
 	if (result != IDEVICE_ACTIVATION_E_SUCCESS) {
 		goto cleanup;
@@ -1073,14 +1351,13 @@ IDEVICE_ACTIVATION_API idevice_activation_error_t idevice_activation_send_reques
 	result = IDEVICE_ACTIVATION_E_SUCCESS;
 
 cleanup:
-	if (iter)
-		free(iter);
+	free(iter);
 	if (form)
 		curl_formfree(form);
+	if (slist)
+		curl_slist_free_all(slist);
 	if (handle)
 		curl_easy_cleanup(handle);
 
-	curl_global_cleanup();
-
 	return result;
 }
